import os

import numpy as np
import xarray as xr


class AnalogueFilter:
    warned = False

    def __init__(self, model_name: str, output_frequency: float = 4e9):
        if not AnalogueFilter.warned:
            print(f"WARNING: Analogue output of simulated {model_name} can differ "
                  "significantly from real hardware!")
            AnalogueFilter.warned = True

        pulse_response_dir = os.path.dirname(__file__)
        if model_name == "QCM":
            fname = pulse_response_dir + "/Qblox_QCM_pulse_response.hdf5"
            amplitude_correction = 1.003
            self.min_vstep = 5.0/2**16  # 0.076 mV
        else:
            raise Exception(f"Uknown model_name {model_name}")

        pulse_response = xr.open_dataset(fname, engine='h5netcdf')
        t_response = pulse_response.coords['t'].data
        self.pulse_response = pulse_response['y'].data * amplitude_correction
        sr = round(1/(t_response[1]-t_response[0]))
        sub_sample = int(round(sr / (output_frequency*1e-9)))
        self.sr = sr // sub_sample
        self.pulse_response = self.pulse_response[::sub_sample]

        self.n_before = round(-t_response[0] * self.sr)
        self.n_after = len(self.pulse_response) - self.n_before - 1

    def quantize_amplitude(self, wave):
        return np.round(wave/self.min_vstep) * self.min_vstep

    def get_awg_output(self, t, samples, analogue_shift: float | None = 0.0):
        samples = self.quantize_amplitude(samples)
        if t[1] - t[0] != 1:
            raise Exception(f"Expecting input with 1 ns sample period. Got {t[1]-t[0]}")
        t = np.linspace(t[0], t[-1]+1, len(t)*self.sr, endpoint=False)
        d = np.zeros(len(samples)*self.sr)
        d[::self.sr] = samples

        d = np.convolve(d, self.pulse_response)
        return t+analogue_shift, d[self.n_before: -self.n_after]



from numbers import Number
from packaging.version import Version
from typing import Any

import pyqtgraph as pg
from qtpy import QtWidgets

from q1simulator import Q1Simulator
from q1simulator.qblox_version import qblox_version


# default colors cycle: see matplotlib CN colors.
color_cycle = [
    '#1f77b4',  # muted blue
    '#ff7f0e',  # safety orange
    '#2ca02c',  # cooked asparagus green
    '#d62728',  # brick red
    '#9467bd',  # muted purple
    '#8c564b',  # chestnut brown
    '#e377c2',  # raspberry yogurt pink
    '#7f7f7f',  # middle gray
    '#bcbd22',  # curry yellow-green
    '#17becf'   # blue-teal
]
color_list = [pg.mkColor(cl) for cl in color_cycle]

_sim_counter = 0


class PlotWindow(QtWidgets.QMainWindow):

    _windows = []

    def __init__(
        self,
    ):
        super().__init__()
        self._n_plots = 0
        self.setWindowTitle("Simulated output")
        self.resize(800, 600)
        self.create_ui()
        self.show()

        # add to static list to avoid garbage collection.
        PlotWindow._windows.append(self)

    def create_ui(self):
        content = QtWidgets.QWidget()

        pw = pg.PlotWidget()
        self._plot_widget = pw
        pw.getAxis('bottom').setLabel("time", "ns")
        pw.getAxis('left').setLabel("output", "V")
        pw.addLegend()

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self._plot_widget, 1)

        content.setLayout(layout)
        self.setCentralWidget(content)

    def plot(self, x, y, style: str = "", label: str = None):
        color = color_list[self._n_plots % len(color_list)]
        pw = self._plot_widget
        pw.plot(x, y, name=label, pen=color)
        self._n_plots += 1

    def closeEvent(self, event):
        PlotWindow._windows.remove(self)


def plot_simulation(
        path: str,
        config: dict[str, Any],
        channels: list[str],
        min_time: int | None = None,
        max_time: int | None = None,
        analogue_filter: bool = False,
        max_core_cycles: int = 10_000_000,
        render_repetitions: bool = False,
        skip_wait_sync: bool = True,
        analogue_output_frequency: float = 4e9,
        ):
    global _sim_counter

    max_render_time = max_time if max_time else 2e6

    # create simulator
    sim = Q1Simulator(
        f"Q1Viewer_{_sim_counter}",
        n_sequencers=len(channels),
        sim_type="Viewer")

    sim.config("max_render_time", max_render_time)
    sim.config("max_core_cycles", max_core_cycles)
    sim.config('skip_loops', () if render_repetitions else ("_start", ))
    sim.config('skip_wait_sync', skip_wait_sync)
    # sim.config('trace', True)
    sim.ignore_triggers = True

    _sim_counter += 1

    # config channels and  load programs
    for i, ch_name in enumerate(channels):
        ch_conf = config[ch_name]

        sequencer = sim.sequencers[i]
        sequencer.label = ch_name
        f_nco = ch_conf["nco"]
        if f_nco is None:
            sequencer.mod_en_awg(False)
        else:
            sequencer.mod_en_awg(True)
            sequencer.nco_freq(f_nco)

        for ch in ch_conf["out_channels"]:
            out_path = ch % 2
            sequencer.parameters[f'connect_out{ch}'].set('I' if out_path == 0 else 'Q')

        sequencer.sequence(path + "/" + ch_conf["sequence"])
        sequencer.sync_en(True)

        sim.arm_sequencer(i)

    # run
    sim.start_sequencer()

    for i, ch_name in enumerate(channels):
        status = sim.get_sequencer_status(i)
        print(f"State {ch_name}: {status.status} - {status.state}, {[str(flag) for flag in status.info_flags]}")
        if status.warn_flags or status.err_flags:
            print(f"  warnings: {[str(flag) for flag in status.warn_flags]}, "
                  f"errors: {[str(flag) for flag in status.err_flags]}, log: {status.log}")

    # plot
    output = sim.get_output(
        t_min=min_time,
        t_max=max_time,
        analogue_filter=analogue_filter,
        output_frequency=analogue_output_frequency,
        )

    pw = PlotWindow()

    for name, data in output.items():
        t, out = data
        if isinstance(t, Number):
            pw.plot(out, label=name)
        elif len(name) > 4 and name[-3:-1] == '-M':
            # marker output
            pw.plot(t, out, ":", label=name)
        else:
            pw.plot(t, out, label=name)

    # sim.print_acquisitions()

    sim.close()

    return pw



import logging
import os

from qcodes import Instrument
from qblox_instruments import Cluster
from .cluster import Cluster as SimCluster


logger = logging.getLogger(__name__)


class Q1Plotter:
    def __init__(self,
                 cluster: Cluster,
                 max_render_time=2e6,
                 max_core_cycles=1e7,
                 skip_loops=("_start", ),
                 skip_wait_sync=True,
                 acq_trigger_value=0,
                 ):

        self._cluster = cluster
        self._simulator = SimCluster(self._get_name(), self._get_module_types())
        self.max_render_time = max_render_time
        self.max_core_cycles = max_core_cycles
        self.skip_loops = skip_loops
        self.skip_wait_sync = skip_wait_sync
        self.acq_trigger_value = acq_trigger_value

        self.load()

    @property
    def max_render_time(self):
        """Maximum time the output is calculated."""
        return self._max_render_time

    @max_render_time.setter
    def max_render_time(self, value):
        self._max_render_time = value
        self._simulator.config('max_render_time', self._max_render_time)

    @property
    def max_core_cycles(self):
        """Maximum number of FPGA clock simulated per sequencer."""
        return self._max_core_cycles

    @max_core_cycles.setter
    def max_core_cycles(self, value):
        self._max_core_cycles = value
        self._simulator.config("max_core_cycles", self._max_core_cycles)

    @property
    def skip_loops(self) -> tuple[str, ...]:
        """Labels of loops to skip.
        Setting this to ("_start", ) will skip the repetitions of Q1Pulse sequences.
        """
        return self._skip_loops

    @skip_loops.setter
    def skip_loops(self, value):
        if value is None:
            value = ()
        self._skip_loops = value
        self._simulator.config("skip_loops", self._skip_loops)

    @property
    def skip_wait_sync(self):
        """If True the time after `wait_sync` is not added to the output."""
        return self._skip_wait_sync

    @skip_wait_sync.setter
    def skip_wait_sync(self, value):
        self._skip_wait_sync = value
        self._simulator.config('skip_wait_sync', self._skip_wait_sync)

    @property
    def acq_trigger_value(self) -> int | None:
        """The value used for every acquisition trigger in the sequence.
        Allowed values: None, 0, 1.

        If None the acquired data is used to determine the trigger value.
        """
        return self._acq_trigger_value

    @acq_trigger_value.setter
    def acq_trigger_value(self, value):
        self._acq_trigger_value = value
        self._simulator.config('acq_trigger_value', self._acq_trigger_value)

    @staticmethod
    def _get_name():
        counter = 1
        name = f'Q1Plotter_{counter}'
        while Instrument.exist(name):
            counter += 1
            name = f'Q1Plotter_{counter}'
        return name

    def _get_module_types(self):
        modules = {}
        for slot, module in enumerate(self._cluster.modules, 1):
            if module.present():
                if module.is_qcm_type:
                    sim_type = "QCM"
                elif module.is_qrm_type:
                    sim_type = "QRM"
                else:
                    logger.warning(f"unknown module type {module.module_type} in slot {slot}")
                    continue
                if module.is_rf_type:
                    sim_type += "_RF"
                modules[slot] = sim_type
        return modules

    def load(self):
        """Loads program and settings from the cluster and generates the output."""
        for slot, module in enumerate(self._simulator.modules, 1):
            if module.present():
                cluster_mod = self._cluster.modules[slot-1]
                module.label = cluster_mod.label
                for iseq in range(6):
                    cluster_seq = cluster_mod.sequencers[iseq]
                    sim_seq = module.sequencers[iseq]
                    enabled = self._copy_settings(cluster_seq, sim_seq, module.is_qcm_type, module.is_rf_type)
                    if enabled:
                        print(f"Generating output for slot {slot}, sequencer {iseq}: {sim_seq.label}")
                        module.arm_sequencer(iseq)

        self._simulator.start_sequencer()

    def regenerate(self):
        """Runs the program again to generate the output."""
        self._simulator.start_sequencer()

    def plot(self,
             t_min: float | None = None,
             t_max: float | None = None,
             channels: list[str] | list[int] | None = None,
             modules: list[int] | None = None,
             create_figure: bool | str = True,
             analogue_filter: bool = False,
             analogue_output_frequency: float = 4e9,
             ):
        """Plots the simulated output of the cluster.

        Args:
            t_min: minimum time in the plot.
            t_max: maximum time in the plot.
            channels: If not None specifies the channels to plot by name or sequencer number.
            modules: If not None specifies the modules to plot by slot number.
            create_figure:
                If True create a new figure.
                If False only pyplot.plot() is called without creating figure or setting axis labels.
                If "modules" creates a new figure per module.
            analogue_filter: plot result after applying (estimated) analog filter.
            analogue_output_frequency: sample rate of analogue output
        """
        end_time = self._simulator.get_simulation_end_time()
        if t_max is not None and t_max > self.max_render_time:
            print(f"The output has only been computed till t={self.max_render_time}. "
                  "Set `max_render_time` and call `regenerate()` to get more data.")
        if t_min is not None and t_min > end_time:
            print(f"The simulation ended at t={end_time}")

        self._simulator.plot(
            t_min=t_min,
            t_max=t_max,
            channels=channels,
            modules=modules,
            create_figure=create_figure,
            analogue_filter=analogue_filter,
            analogue_output_frequency=analogue_output_frequency,
            )

    @staticmethod
    def _copy_settings(cluster_seq, sim_seq, is_qcm, is_rf):
        sim_seq.sync_en(cluster_seq.sync_en.cache())
        if not sim_seq.sync_en():
            return False
        sequence = cluster_seq.sequence.cache()
        if not isinstance(sequence, dict):
            if not os.path.exists(sequence):
                raise Exception(f"Cannot load sequence {sequence}")
        sim_seq.sequence(sequence)
        sim_seq.label = cluster_seq.label

        param_names = [
                "mod_en_awg",
                "nco_freq",
                'marker_ovr_en',
                'marker_ovr_value',
                'nco_phase_offs',
                'gain_awg_path0',
                'gain_awg_path1',
                'offset_awg_path0',
                'offset_awg_path1',
                ]

        for i in range(1, 16):
            param_names += [
                f'trigger{i}_count_threshold',
                f'trigger{i}_threshold_invert',
                ]

        n_out_ch = 4 if is_qcm else 2
        if is_rf:
            n_out_ch //= 2
        for ch in range(n_out_ch):
            param_names += [f"connect_out{ch}"]

        if not is_qcm:
            param_names += [
                "demod_en_acq",
                "integration_length_acq",
                'thresholded_acq_rotation',
                'thresholded_acq_threshold',
                'thresholded_acq_trigger_en',
                'thresholded_acq_trigger_address',
                'thresholded_acq_trigger_invert',
                ]

        for param_name in param_names:
            value = cluster_seq.parameters[param_name].cache()
            if value is not None:
                sim_seq.parameters[param_name].set(value)

        return True



import json
from qtpy import QtWidgets, QtCore

from .program_plotter import plot_simulation


class ProgramDetailsWidget(QtWidgets.QWidget):
    def __init__(self,
                 parent: QtWidgets.QWidget,
                 ):
        super().__init__(parent)
        self._prog_config = None
        self._path = None

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self._table = self.create_table()
        layout.addWidget(self._table, 1)
        layout.addLayout(self.create_controls())

        self._plot_btn.clicked.connect(self.plot)

    def create_table(self):
        table = QtWidgets.QTreeWidget()
        table.setMinimumHeight(400)
        table.setMinimumWidth(540)
        table.setColumnCount(7)
        table.setColumnWidth(0, 100)
        table.setColumnWidth(1, 100)
        table.setColumnWidth(2, 50)
        table.setColumnWidth(3, 50)
        table.setColumnWidth(4, 50)
        table.setColumnWidth(5, 90)
        table.setColumnWidth(6, 80)
        table.setHeaderLabels([
            "Channel",
            "Module",
            "Paths",
            "Out",
            "In",
            "NCO",
            "Duration",
            ])
        table.setRootIsDecorated(False)
        return table

    def create_controls(self):
        min_time = QtWidgets.QSpinBox()
        min_time.setRange(0, 100_000_000)
        min_time.setSingleStep(1000)
        self._min_time_spin = min_time
        max_time = QtWidgets.QSpinBox()
        max_time.setSpecialValueText("Duration")
        max_time.setRange(0, 100_000_000)
        max_time.setSingleStep(1000)
        self._max_time_spin = max_time

        self._analogue_cb = QtWidgets.QCheckBox("Analogue filter")
        self._plot_btn = QtWidgets.QPushButton("Plot")

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(QtWidgets.QLabel("t min."))
        layout.addWidget(min_time)
        layout.addWidget(QtWidgets.QLabel("t max."))
        layout.addWidget(max_time)
        layout.addWidget(self._analogue_cb)
        layout.addWidget(self._plot_btn)

        return layout

    def show_details(self, path):
        table = self._table
        table.clear()
        self._prog_config = None
        try:
            with open(path + "/q1program.json", "r") as fp:
                data = json.load(fp)
            self._path = path
            self._prog_config = data
            duration = -1
            rows = []
            for ch_name, settings in data.items():
                f_nco = settings["nco"]
                seq_duration = settings.get("duration")
                if seq_duration is not None:
                    duration = max(seq_duration, duration)
                    duration_text = f"{seq_duration} ns"
                else:
                    duration_text = "--"
                q1asm = settings.get("sequence")
                if not q1asm:
                    duration_text = "no sequence"

                item = QtWidgets.QTreeWidgetItem(table)
                item.setText(0, ch_name)
                item.setData(0, QtCore.Qt.CheckStateRole, QtCore.Qt.Checked if q1asm else QtCore.Qt.Unchecked)

                item.setText(1, settings["module"])
                item.setText(2, str(settings["paths"]))
                item.setText(3, str(settings["out_channels"]))
                item.setText(4, str(settings.get("in_channels", "")))
                item.setText(5, f"{f_nco/1e6} MHz" if f_nco is not None else "")
                item.setText(6, duration_text)
                rows.append(item)
            table.insertTopLevelItems(0, rows)

            # Show duration

        except FileNotFoundError:
            print("'program_config.json' not found")

    def plot(self):
        if not self._prog_config:
            print("no program")
            return
        table = self._table
        channels = []
        for i in range(table.topLevelItemCount()):
            item = table.topLevelItem(i)
            if item.checkState(0) == QtCore.Qt.Checked:
                channels.append(item.text(0))

        min_time = self._min_time_spin.value()
        max_time = self._max_time_spin.value()
        if max_time == 0:
            max_time = None

        plot_simulation(
            self._path,
            self._prog_config,
            channels,
            min_time=min_time,
            max_time=max_time,
            analogue_filter=self._analogue_cb.isChecked(),
            analogue_output_frequency=4e9,
            )



import logging
from dataclasses import dataclass

from .q1sequencer import Q1Sequencer


logger = logging.getLogger(__name__)

@dataclass
class SequencerTriggerInfo:
    sequencer: Q1Sequencer
    emitted_triggers: int
    ''' bits for emitted trigger.
    0b00001010 for triggers 2 and 4.
    '''
    used_triggers: int
    ''' bits for all triggers used by sequencer conditions.
    0b00001010 for triggers 2 and 4.
    '''

def get_seq_trigger_info(sequencer):
    if hasattr(sequencer,'thresholded_acq_trigger_en') and sequencer.thresholded_acq_trigger_en():
        emitted_triggers = 1 << (sequencer.thresholded_acq_trigger_address() - 1)
    else:
        emitted_triggers = 0
    return SequencerTriggerInfo(
            sequencer,
            emitted_triggers,
            sequencer.get_used_triggers()
            )

def get_emitted_triggers(sequencers: list[SequencerTriggerInfo]):
    emitted = 0
    for seq in sequencers:
        emitted |= seq.emitted_triggers
    return emitted

def sort_sequencers(sequencers: list[SequencerTriggerInfo]):
    ''' Sorts sequencers putting trigger producers before trigger consumers.

    Sequencers are executed sequentially. A sequencer must be executed after
    all sequencers that emit triggers that it uses.

    Notes:
        * A sequencer may use a trigger not emitted by any sequencer.
        * A sequencer may use a trigger that it also emits.
    '''
    result = []
    unsorted = sequencers.copy()
    while len(unsorted) > 0:
        n_start = len(unsorted)
        to_be_emitted = get_emitted_triggers(unsorted)
        for seq in unsorted.copy():
            if seq.used_triggers & to_be_emitted == 0:
                # sequence does not use triggers that will be emitted after it.
                result.append(seq)
                unsorted.remove(seq)
            elif seq.used_triggers & to_be_emitted & ~seq.emitted_triggers == 0:
                # sequence emits a triggers it uses. Check if there are other sequencers emitting the trigger.
                remaining = unsorted.copy()
                remaining.remove(seq)
                rem_emitted = get_emitted_triggers(remaining)
                if seq.used_triggers & rem_emitted == 0:
                    # OK. It does not use triggers of remaining sequencers.
                    result.append(seq)
                    unsorted.remove(seq)

        if len(unsorted) == n_start:
            raise Exception(f"Sorting sequencers on triggers failed. Unsorted: {unsorted}")

    return result
